@@ -7,6 +7,7 @@ import copy
 import json
 from collections import namedtuple
 import itertools
+from numbers import Number
 from math import isnan
 import traceback
 
@@ -20,16 +21,38 @@ MultiPolygon = namedtuple('MultiPolygon', ['coordinates', 'crs'])
 GeometryCollection = namedtuple('GeometryCollection', ['geometries'])
 
 
+class ExtendedJSONEncoder(json.JSONEncoder):
+    """ Numpy-specific numbers prior to 1.9 don't inherit from Python numeric
+    ABCs. This class is a hack to coerce numpy values into Python types for
+    JSON serialization. """
+
+    def default(self, o):
+        try:
+            if o.dtype in ("int8", "int16", "int32", "int64"):
+                return int(o)
+            elif o.dtype in ("float16", "float32", "float64", "float128"):
+                return float(o)
+            elif o.dtype in ("complex64", "complex128", "complex256"):
+                return complex(o)
+            else:
+                raise TypeError("not a recognized type")
+        except (AttributeError, TypeError):
+            return json.JSONEncoder.default(self, o)
+
+
 class GeoJSONWriter(object):
     """ Class for converting guppy objects to GeoJSON strings. Multipoint-based
     opbjects are written as 'Features'.
 
+    CRS defaults to be named. A linked CRS can be used by passing
+    `linkedcrs=True`.
+
     Notes:
     ------
     Does not handle 'FeatureCollection' types (see printFeatureCollection()
     function).
     """
-    def __init__(self, gpobj, **kwargs):
+    def __init__(self, gpobj, linkedcrs=False, **kwargs):
         type_equiv = {'Point'       : 'Point',
                       'Multipoint'  : 'MultiPoint',
                       'Line'        : 'LineString',
@@ -38,16 +61,14 @@ class GeoJSONWriter(object):
         if gpobj._geotype in type_equiv:
             self.typestr = type_equiv[gpobj._geotype]
         else:
-            raise NotImplementedError('input object not a recognizes guppy '
-                                      'type')
-        crs = kwargs.get('crs', None)
-        crs_fmt = kwargs.get('crs_fmt', None)
+            raise TypeError('Input object not a recognized geometry')
+        crs = gpobj._crs
         bbox = kwargs.get('bbox', None)
 
         self.gpobj = gpobj
         self.supobj = {}
         if crs is not None:
-            self.add_named_crs(crs, crs_fmt)
+            self.add_crs(crs, linkedcrs=linkedcrs)
         if self.typestr != 'Point':
             self.supobj['type'] = 'Feature'
             self.add_bbox(bbox)
@@ -57,27 +78,30 @@ class GeoJSONWriter(object):
             self.add_coordinates()
         return
 
-    def add_named_crs(self, crs, fmt='epsg'):
+    def add_crs(self, crs, linkedcrs, **kw):
         """ Tag the JSON object with coordinate reference system metadata using
-        a named CRS. `fmt` indicates the format of the named CRS argument, and
-        may be one of ('epsg', 'ogc_crs_urn').
-        """
-        if fmt == 'epsg':
-            href = ('http://spatialreference.org/ref/epsg/{0}/'
-                    'proj4/'.format(crs))
-            islinked = True
-        elif fmt == 'ogc_crs_urn':
-            crs_str = crs
-            islinked = False
-        else:
-            raise NotImplementedError('CRS fmt {0} not recognized'.format(fmt))
+        a CRS instance, which is expected the have the interface of crs.CRS.
 
-        if islinked:
+        If True, `linkedcrs` indicates a web-referenced linked CRS. If False,
+        the CRS is taked to be named.
+
+        In the case of a linked CRS, the link address and type can be specified
+        using the `href` and `linktype` keyword arguments. If not supplied,
+        then it is assumed that crs.urn contains the *href* and crs.type
+        contains the *type*.
+
+        For more details, see the GeoJSON specification at:
+        http://geojson.org/geojson-spec.html#coordinate-reference-system-objects
+        """
+        if linkedcrs:
+            href = kw.get('href', crs.urn)
+            linktype = kw.get('linktype', crs.type)
             self.supobj['crs'] = {'type': 'link',
-                                  'properties': {'href': href, 'type': 'proj4'}}
+                                  'properties': {'href': href, 'type': linktype}}
+
         else:
             self.supobj['crs'] = {'type': 'name',
-                                  'properties': {'name': crs_str}}
+                                  'properties': {'name': crs.urn}}
         return
 
     def add_bbox(self, bbox=None):
@@ -91,7 +115,7 @@ class GeoJSONWriter(object):
             else:
                 raise AttributeError("Type {0} has no 'get_extents' "
                                      "method".format(type(self.gpobj)))
-        self.supobj['bbox'] = {'bbox'    :   bbox}
+        self.supobj['bbox'] = bbox
         return
 
     def add_geometry(self):
@@ -134,8 +158,8 @@ class GeoJSONWriter(object):
             else:
                 data = {'point_data': self.gpobj.data}
             for key in data:
-                target['properties'][key] = map(lambda a: a if not isnan(a) else None,
-                                                data.getfield(key))
+                nan2none = lambda a: None if isinstance(a, Number) and isnan(a) else a
+                target['properties'][key] = list(map(nan2none, data.getfield(key)))
         return
 
     def add_id(self, target=None):
@@ -147,12 +171,12 @@ class GeoJSONWriter(object):
 
     def print_json(self):
         """ Print GeoJSON representation. """
-        return json.dumps(self.supobj, indent=2)
+        return json.dumps(self.supobj, indent=2, cls=ExtendedJSONEncoder)
 
     def write_json(self, fout):
         """ Dump internal dict-object to JSON using the builtin `json` module.
         """
-        json.dump(self.supobj, fout, indent=2)
+        json.dump(self.supobj, fout, indent=2, cls=ExtendedJSONEncoder)
         return
 
 
@@ -176,21 +200,19 @@ class GeoJSONReader(object):
     -----------------------------------------------------
      """
 
-    def __init__(self, finput=None):
+    def __init__(self, finput):
         """ Create a reader-object for a GeoJSON-containing file or StreamIO
         object. Use as::
 
             with open(`fnm`, 'r') as f:
                 reader = GeoJSONReader(f)
         """
-        if finput is not None:
-            if not hasattr(finput, 'read'):
-                with open(finput) as f:
-                    self.jsondict = json.load(f)
-            else:
-                self.jsondict = json.load(finput)
+        if not hasattr(finput, 'read'):
+            with open(finput) as f:
+                self.jsondict = json.load(f)
         else:
-            self.jsondict = None
+            self.jsondict = json.load(finput)
+        self.crs = self.getcrs()
         return
 
     def _walk(self, dic, geotype):
@@ -217,8 +239,30 @@ class GeoJSONReader(object):
                     return True
         return False
 
+    def _getproperties(self):
+        """ Read feature properties and return a two dictionaries:
+            - one representing singleton values ("properties")
+            - one representing per-vertex values ("data")
+        """
+        raise NotImplementedError
+        return {}, {}
+
+    def getcrs(self):
+        """ Look in the top level for a `crs` field. Return a tuple that is
+        either (name, _), (href, type), or (_, _) for a named, linked, or
+        undefined CRS, respectively. """
+        crs = self.jsondict.get("crs", None)
+        if crs is not None:
+            if "name" in crs["properties"]:
+                res = (crs["properties"]["name"], None)
+            else:
+                res = (crs["properties"]["href"], crs["properties"]["type"])
+        else:
+            res = (None, None)
+        return res
+
     def pull_features(self):
-        """ Find and return all Feature objects """
+        """ Find and return all Feature objects from a FeatureCollection """
         jsonfeats = [obj for obj in self._walk(self.jsondict, 'Feature')]
 
         featuretype = lambda a: a.get('geometry', {}).get('type', None)
@@ -227,6 +271,8 @@ class GeoJSONReader(object):
         isline = lambda a: featuretype(a) in ('LineString', 'MultiLineString')
         ispolygon = lambda a: featuretype(a) in ('Polygon', 'MultiPolygon')
 
+        # This is an idiotic algorithm - pre-mapping the types isn't saving
+        # time because each feature needs to be iterated through anyway
         points = filter(ispoint, jsonfeats)
         multipoints = filter(ismultipoint, jsonfeats)
         lines = filter(isline, jsonfeats)
@@ -234,13 +280,17 @@ class GeoJSONReader(object):
 
         features = []
         for feat in points:
-            features.append((self.pull_points(feat), feat['properties'], feat.get('id', None)))
+            features.append((self.pull_points(feat)[0], feat['properties'],
+                             feat.get('id', None)))
         for feat in multipoints:
-            features.append((self.pull_multipoints(feat), feat['properties'], feat.get('id', None)))
+            features.append((self.pull_multipoints(feat)[0], feat['properties'],
+                             feat.get('id', None)))
         for feat in lines:
-            features.append((self.pull_lines(feat), feat['properties'], feat.get('id', None)))
+            features.append((self.pull_lines(feat)[0], feat['properties'],
+                             feat.get('id', None)))
         for feat in polygons:
-            features.append((self.pull_polygons(feat), feat['properties'], feat.get('id', None)))
+            features.append((self.pull_polygons(feat)[0], feat['properties'],
+                             feat.get('id', None)))
 
         return features
 
@@ -252,7 +302,7 @@ class GeoJSONReader(object):
         jsonpoints = self._walk(dic, 'Point')
         points = []
         for point in jsonpoints:
-            points.append(Point(point['coordinates'], None))
+            points.append(Point(point['coordinates'], self.crs))
         return points
 
     def pull_multipoints(self, dic=None):
@@ -263,7 +313,7 @@ class GeoJSONReader(object):
         jsonmultipoints = self._walk(dic, 'MultiPoint')
         multipoints = []
         for jsonmultipoint in jsonmultipoints:
-            multipoints.append(MultiPoint(jsonmultipoint['coordinates'], None))
+            multipoints.append(MultiPoint(jsonmultipoint['coordinates'], self.crs))
         return multipoints
 
     def pull_lines(self, dic=None):
@@ -275,10 +325,10 @@ class GeoJSONReader(object):
         jsonmultilines = self._walk(dic, 'MultiLineString')
         lines = []
         for jsonline in jsonlines:
-            lines.append(LineString(jsonline['coordinates'], None))
+            lines.append(LineString(jsonline['coordinates'], self.crs))
         for jsonmultiline in jsonmultilines:
             for vertices in jsonmultiline['coordinates']:
-                lines.append(MultiLineString(vertices, None))
+                lines.append(MultiLineString(vertices, self.crs))
         return lines
 
     def pull_polygons(self, dic=None):
@@ -290,9 +340,9 @@ class GeoJSONReader(object):
         jsonmultipolygons = self._walk(dic, 'MultiPolygon')
         polygons = []
         for jsonpolygon in jsonpolygons:
-            polygons.append(Polygon(jsonpolygon['coordinates'], None))
+            polygons.append(Polygon(jsonpolygon['coordinates'], self.crs))
         for jsonmultipolygon in jsonmultipolygons:
-            polygons.append(Polygon(jsonmultipolygon['coordinates'], None))
+            polygons.append(Polygon(jsonmultipolygon['coordinates'], self.crs))
         return polygons
 
     def iter_geometries(self):
@@ -303,9 +353,6 @@ class GeoJSONReader(object):
                                   self.pull_polygons())
         return itergeo
 
-    def list_features(self):
-        """ List the features present. """
-        raise NotImplementedError
 
 def list_rec(A):
     """ Recursively convert nested iterables to nested lists """
@@ -343,17 +390,11 @@ def geojson2csv(fin, fout):
     X = []
     Y = []
     PROP = []
-    try:
-        for feat in jsdat['features']:
-            x, y = feat['geometry']['coordinates']
-            X.append(x)
-            Y.append(y)
-            PROP.append(feat['properties'])
-    except KeyError:
-        sys.stderr.write('invalid geojson schema')
-        traceback.print_exc()
-    except:
-        traceback.print_exc()
+    for feat in jsdat['features']:
+        x, y = feat['geometry']['coordinates']
+        X.append(x)
+        Y.append(y)
+        PROP.append(feat['properties'])
 
     # Find subset of properties that exist in all features
     if len(PROP) > 1:
