@@ -51,7 +51,7 @@ class Geometry(object):
     def _distance(self, pos0, pos1):
         """ Generic method for calculating distance between positions that
         respects CRS """
-        if self._crs.type == "projected":
+        if self._crs == crs.CARTESIAN:
             dist = _vecgeo.distance((pos0.x, pos0.y), (pos1.x, pos1.y))
         elif PYPROJ:
             _, _, dist = geod.inv(pos0.x, pos0.y, pos1.x, pos1.y, radians=False)
@@ -104,7 +104,7 @@ class Point(Geometry):
 
     def __eq__(self, other):
         if hasattr(other, "vertex"):
-            return (self.vertex == other.vertex) and \
+            return (tuple(self.vertex) == tuple(other.vertex)) and \
                    (self.data == other.data) and \
                    (self.properties == other.properties)
         else:
@@ -143,14 +143,15 @@ class Point(Geometry):
             return (self.x, self.y)
 
     def azimuth(self, other, spherical=False):
-        """ Returns the azimuth from self to other in radians. Returns
-        None if points are coincident.
-        """
+        """ Returns the compass azimuth from self to other in radians (i.e.
+        clockwise, with north at 0). Returns NaN if points are coincident. """
+
+        if self.coordsxy() == other.coordsxy():
+            return np.nan
 
         if self._crs == crs.LONLAT and other._crs == crs.LONLAT:
             if not PYPROJ:
-                raise CRSError("Azimuth computation on LONLAT grids "
-                               "requires pyproj")
+                raise CRSError("Azimuth on LONLAT grids requires pyproj")
             az1, _, _ = geod.inv(self.x, self.y, other.x, other.y)
             return az1 * math.pi / 180.0
 
@@ -158,19 +159,25 @@ class Point(Geometry):
             dx = other.x - self.x
             dy = other.y - self.y
 
-            if dx == 0.0:
-                if dy > 0.0:
-                    return 0.0
-                elif dy < 0.0:
-                    return math.pi
+            if dx > 0:
+                if dy > 0:
+                    return math.atan(dx/dy)
+                elif dy < 0:
+                    return math.pi - math.atan(-dx/dy)
                 else:
-                    return None
-
-            elif dy >= 0.0:
-                return math.atan(dy / dx)
-
+                    return 0.5*math.pi
+            elif dx < 0:
+                if dy > 0:
+                    return 2*math.pi - math.atan(-dx/dy)
+                elif dy < 0:
+                    return math.pi + math.atan(dx/dy)
+                else:
+                    return 1.5*math.pi
             else:
-                return math.atan(dy / dx) + math.pi
+                if dy > 0:
+                    return 0.0
+                else:
+                    return math.pi
 
         else:
             raise CRSError("Azimuth undefined for points in CRS {0} and "
@@ -192,7 +199,6 @@ class Point(Geometry):
         else:
             return math.sqrt(flat_dist**2. + (self.z-other.z)**2.)
 
-
     def greatcircle(self, other):
         """ Return the great circle distance between two geographical points. """
         if not PYPROJ:
@@ -240,8 +246,16 @@ class Point(Geometry):
         crs_fmt : format of `crs`; may be one of ('epsg','ogc_crs_urn')
         bbox : an optional bounding box tuple in the form (w,e,s,n)
         """
-        writer = geojson.GeoJSONWriter(self, **kwargs)
-        writer.write_json(f)
+        try:
+            if not hasattr(f, "write"):
+                fobj = open(f, "w")
+            else:
+                fobj = f
+            writer = geojson.GeoJSONWriter(self, **kwargs)
+            writer.write_json(fobj)
+        finally:
+            if not hasattr(f, "write"):
+                fobj.close()
         return writer
 
     def to_shapely(self):
@@ -525,7 +539,8 @@ class MultipointBase(Geometry):
         crs_fmt : format of `crs`; may be one of ('epsg','ogc_crs_urn')
         bbox : an optional bounding box tuple in the form (w,e,s,n)
         """
-        writer = geojson.GeoJSONWriter(self, **kwargs)
+        writer = geojson.GeoJSONWriter(self, crs=self._crs,
+                    crs_fmt="ogc_crs_urn", **kwargs)
         return writer.print_json()
 
     def to_geojson(self, f, **kwargs):
@@ -540,13 +555,26 @@ class MultipointBase(Geometry):
         crs_fmt : format of `crs`; may be one of ('epsg','ogc_crs_urn')
         bbox : an optional bounding box tuple in the form (w,e,s,n)
         """
-        writer = geojson.GeoJSONWriter(self, **kwargs)
-        writer.write_json(f)
+        try:
+            if not hasattr(f, "write"):
+                fobj = open(f, "w")
+            else:
+                fobj = f
+            writer = geojson.GeoJSONWriter(self, crs=self._crs,
+                        crs_fmt="ogc_crs_urn", **kwargs)
+            writer.write_json(fobj)
+        finally:
+            if not hasattr(f, "write"):
+                fobj.close()
         return writer
 
     def to_vtk(self, f, **kwargs):
         """ Write data to an ASCII VTK .vtp file. """
-        vtk.mp2vtp(self, f, **kwargs)
+        if not hasattr(f, "write"):
+            with open(f, "w") as fobj:
+                vtk.mp2vtp(self, fobj, **kwargs)
+        else:
+            vtk.mp2vtp(self, f, **kwargs)
         return
 
     def to_shapefile(self, fstem):
@@ -573,20 +601,54 @@ class Multipoint(MultipointBase):
         point attributes. If `data` is not `None`, then it (or its values) must
         match `vertices` in length.
         """
-        super(Multipoint, self).__init__(vertices,
-                                         data=data, 
-                                         properties=properties,
-                                         **kwargs)
+        if False not in (hasattr(v, "_geotype") and \
+                         v._geotype == "Point" and \
+                         v.rank == vertices[0].rank for v in vertices):
+            points = vertices
+            crs = points[0]._crs
+            if False in (pt._crs == crs for pt in points):
+                raise CRSError("All points must share the same CRS")
+
+            keys = list(points[0].properties.keys())
+            for pt in points[1:]:
+                for key in keys:
+                    if key not in pt.properties:
+                        keys.pop(keys.index(key))
+
+            ptdata = {}
+            for key in keys:
+                ptdata[key] = [pt.properties[key] for pt in points]
+
+            if data is not None:
+                ptdata.update(data)
+            self.data = Metadata(ptdata)
+            self.vertices = [pt.vertex for pt in points]
+            self.rank = vertices[0].rank
+            self.properties = properties
+            self._crs = crs
+
+        else:
+
+            super(Multipoint, self).__init__(vertices,
+                                             data=data, 
+                                             properties=properties,
+                                             **kwargs)
         return
 
     def within_radius(self, pt, radius):
-        """ Return Multipoint of subset of member vertices that are within
-        *radius* of *pt*.
+        """ Return Multipoint of subset that is within *radius* of *pt*.
         """
         distances = self.distances_to(pt)
-        nearidx = [i for i,d in enumerate(distances) if d < radius]
-        subset = self._subset(nearidx)
-        return subset
+        indices = [i for i,d in enumerate(distances) if d <= radius]
+        return self._subset(indices)
+
+    def within_bbox(self, bbox):
+        """ Return Multipoint subset that is within a square boundaing box
+        given by (xmin, xmax, ymin, ymax). """
+        filtbbox = lambda pt: (bbox[0] <= pt.vertex[0] <= bbox[1]) and \
+                              (bbox[2] <= pt.vertex[1] <= bbox[3])
+        indices = [i for (i, pt) in enumerate(self) if filtbbox(pt)]
+        return self._subset(indices)
 
 
 class ConnectedMultipoint(MultipointBase):
@@ -709,6 +771,28 @@ class Line(ConnectedMultipoint):
         """ Returns the distance between the first and last vertex. """
         return Point(self.vertices[0]).distance(Point(self.vertices[-1]))
 
+    def direction(self):
+        """ Returns a vector of the azimuth along a line at each point,
+        weighted by neighbour position. """
+        # The weights are calculated to give greater weight to azimuths to
+        # nearby points
+        # a' = w1 a1 + w2 a2
+        #    = (a1 h2) / (h1 + h2) + (a2 h1) / (h1 + h2)
+        #    = (a1 h2 + a2 h1) / (h1 + h2)
+        azs = np.empty(len(self)-1)
+        hs = np.empty(len(self)-1)
+        pts = [pt for pt in self]
+        for i in range(len(self)-1):
+            azs[i] = pt[i].azimuth(pt[i+1])
+            hs[i] = pt[i].distance(pt[i+1])
+        weighted_azs = np.empty(len(self))
+        weighted_azs[0] = azs[0]
+        weighted_azs[-1] = azs[-1]
+        for i in range(1, len(self)-1):
+            raise NotImplementedError("need to phase unwrap")
+            weighted_azs[i] = (azs[i]*hs[i+1] + azs[i+1]*hs[i]) / (hs[i] + hs[i+1])
+        return weighted_azs
+
     def to_polygon(self):
         """ Returns a polygon. """
         return Polygon(self.vertices)
@@ -748,6 +832,14 @@ class Polygon(ConnectedMultipoint):
         self.subs = subs if subs is not None else []
         return
 
+    def __getitem__(self, key):
+        if isinstance(key, slice):
+            ind = key.indices(len(self))
+            if len(self) != ((ind[1] - ind[0]) // ind[2]):
+                return Line(self.vertices[key], data=self.data[key], 
+                            properties=self.properties, crs=self._crs)
+        return super(Polygon, self).__getitem__(key)
+
     def perimeter(self):
         """ Return the perimeter of the polygon. If there are sub-polygons,
         their perimeters are added recursively. """
@@ -974,6 +1066,28 @@ def sortby(A, B):
     return [i[1] for i in comb]
 
 
+def points_to_multipoint(points):
+    """ Merge *points* into a Multipoint instance. Point properties are stored
+    as Multipoint data. All points must use the same CRS.
+    """
+    crs = points[0]._crs
+    if False in (pt._crs == crs for pt in points):
+        raise CRSError("All points must share the same CRS")
+
+    keys = list(points[0].properties.keys())
+    for pt in points[1:]:
+        for key in keys:
+            if key not in pt.properties:
+                keys.pop(keys.index(key))
+
+    ptdata = {}
+    for key in keys:
+        ptdata[key] = [pt.properties[key] for pt in points]
+
+    vertices = [pt.vertex for pt in points]
+
+    return Multipoint(vertices, data=ptdata, crs=crs)
+
 def tighten(X, Z):
     """ Return a list of corrected measurements from observations of
     topography across a cross-section. The inputs are equal length
