@@ -5,6 +5,20 @@ import shapefile
 from . import guppy
 from . import geojson
 from . import xyfile
+from .. import crs
+
+def _parsegeojsoncrs(crstup):
+    """ From a tuple representing a GeoJSON (name,None) or (href,type) pair,
+    return an appropriate karta crs instance. """
+    if crstup[0] is None:       # use default as defined by spec
+        return crs.LONLAT_WGS84
+    elif crstup[1] is None:     # named CRS
+        for c in crs.crslist:
+            if c.urn == crstup[0]:
+                return c
+        return crs.CRS("unknown", "unknown", crstup[0])
+    else:                       # linked CRS
+        return crs.CRS("unknown", crstup[1], crstup[0])
 
 def read_geojson(f):
     """ Read a GeoJSON object file and return a list of geometries """
@@ -12,29 +26,64 @@ def read_geojson(f):
     geoms = R.iter_geometries()
     gplist = []
     for geom in geoms:
+        coordsys = _parsegeojsoncrs(geom.crs)
         if isinstance(geom, geojson.Point):
-            gplist.append(guppy.Point(geom.coordinates))
+            gplist.append(guppy.Point(geom.coordinates, crs=coordsys))
         elif isinstance(geom, geojson.MultiPoint):
-            gplist.append(guppy.Multipoint(geom.coordinates))
+            gplist.append(guppy.Multipoint(geom.coordinates, crs=coordsys))
         elif isinstance(geom, geojson.LineString):
-            gplist.append(guppy.Line(geom.coordinates))
+            gplist.append(guppy.Line(geom.coordinates, crs=coordsys))
         elif isinstance(geom, geojson.Polygon):
             gplist.append(guppy.Polygon(geom.coordinates[0],
-                                        subs=geom.coordinates[1:]))
+                                        subs=geom.coordinates[1:],
+                                        crs=coordsys))
     return gplist
 
 def read_geojson_features(f):
     """ Read a GeoJSON object file and return a list of features """
     R = geojson.GeoJSONReader(f)
-    feats = R.pull_features()
-    gplist = []
-    raise NotImplementedError
-    for feat in feats:
+    features = R.pull_features()
+    print()
+    geoms = []
+    coordsys = _parsegeojsoncrs(R.getcrs())
+    for (geom, properties, id) in features:
 
-        # [...]
-        pass
+        if isinstance(geom, geojson.Point):
+            (p, d) = _geojson_properties2guppy(properties, 1)
+            geoms.append(guppy.Point(geom.coordinates, properties=p, data=d,
+                                     crs=coordsys))
 
-    return gplist
+        elif isinstance(geom, geojson.MultiPoint):
+            (p, d) = _geojson_properties2guppy(properties, len(geom.coordinates))
+            geoms.append(guppy.Multipoint(geom.coordinates, properties=p, data=d,
+                                          crs=coordsys))
+
+        elif isinstance(geom, geojson.LineString):
+            (p, d) = _geojson_properties2guppy(properties, len(geom.coordinates))
+            geoms.append(guppy.Line(geom.coordinates, properties=p, data=d,
+                                    crs=coordsys))
+
+        elif isinstance(geom, geojson.Polygon):
+            (p, d) = _geojson_properties2guppy(properties, len(geom.coordinates))
+            geoms.append(guppy.Polygon(geom.coordinates[0], properties=p, data=d,
+                                       subs=geom.coordinates[1:],
+                                       crs=coordsys))
+    return geoms
+
+def _geojson_properties2guppy(properties, n):
+    """ Takes a dictionary (derived from a GeoJSON properties object) and
+    divides it into singleton properties and *n*-degree data. """
+    props = {}
+    data = {}
+    for (key, value) in properties.items():
+        if isinstance(value, list) or isinstance(value, tuple):
+            if len(value) == n:
+                data[key] = value
+            else:
+                raise ValueError("properties must be singleton or per-vertex")
+        else:
+            props[key] = value
+    return props, data
 
 
 def shape2point(shape):
